from typing import TYPE_CHECKING, Iterator, List

from angrmanagement.plugins import BasePlugin
from angrmanagement.ui.widgets.qinst_annotation import QInstructionAnnotation, QPassthroughCount

if TYPE_CHECKING:
    from angr.sim_manager import SimulationManager

    from angrmanagement.ui.widgets.qblock import QBlock


class SamplePlugin(BasePlugin):
    def __init__(self, workspace):
        super().__init__(workspace)

        workspace.main_instance.register_container("bookmarks", list, List[int], "Bookmarked addresses")

    MENU_BUTTONS = ("Add Bookmark",)

    def build_context_menu_functions(self, funcs):  # pylint: disable=unused-argument
        yield ("owo", [("uwu", lambda: None), ("o_O", lambda: None)])

    def step_callback(self, simgr: "SimulationManager"):
        print("Active States: %s" % simgr)

    def build_qblock_annotations(self, qblock: "QBlock") -> Iterator[QInstructionAnnotation]:
        return [QPassthroughCount(qblock.addr, "entry")]



from PySide6.QtWidgets import QGraphicsItem


class QCachedGraphicsItem(QGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._cached_bounding_rect = None
        self._cached_device_pixel_ratio = None

    def clear_cache(self):
        self.prepareGeometryChange()
        self._cached_bounding_rect = None
        self._cached_device_pixel_ratio = None

    def refresh(self):
        pass

    @property
    def width(self):
        return self.boundingRect().width()

    @property
    def height(self):
        return self.boundingRect().height()

    def recalculate_size(self):
        self.prepareGeometryChange()
        self._cached_device_pixel_ratio = None
        self._cached_bounding_rect = self._boundingRect()

    def boundingRect(self):
        if self._cached_bounding_rect is None:
            self._cached_bounding_rect = self._boundingRect()
        return self._cached_bounding_rect

    def _boundingRect(self):
        raise NotImplementedError

    def _boundingRectAdjusted(self):
        # adjust according to devicePixelRatioF
        return self._boundingRect()



class JumpHistory:
    """
    A class to store the navigation history of a reversing session. Typically found at DisassemblyView._jump_history
    or CodeView.jump_history. Maintains a list of addresses through which the user can navigate forwards and backwards.
    """

    def __init__(self):
        self._history = []
        self._pos = -1

    @property
    def history(self):
        return self._history

    @property
    def pos(self):
        return self._pos

    @property
    def current(self):
        if len(self._history):
            return self._history[self._pos]
        else:
            return None

    def __len__(self):
        return len(self._history)

    def jump_to(self, addr):
        if self._pos != len(self._history) - 1:
            self.trim()

        if not self._history or self._history[-1] != addr:
            self._history.append(addr)
            self._pos = len(self._history) - 1

    def record_address(self, addr):
        if 0 <= self._pos < len(self._history):
            self._history[self._pos] = addr
        else:
            self.jump_to(addr)

    def trim(self):
        self._history = self._history[: self._pos + 1]

    def backtrack(self):
        if self._pos > 0:
            self._pos -= 1

        if not 0 <= self._pos < len(self._history):
            return None
        else:
            return self._history[self._pos]

    def forwardstep(self):
        if self._pos < len(self._history) - 1:
            self._pos += 1

        if self._pos < len(self._history):
            return self._history[self._pos]
        else:
            return None

    def step_position(self, pos: int):
        if -1 < pos < len(self._history):
            self._pos = pos
        return self._history[self._pos]



class ToolbarAction:
    def __init__(self, icon, name, tooltip, triggered, checkable=False, shortcut=None):
        self.icon = icon
        self.name = name
        self.tooltip = tooltip
        self.triggered = triggered
        self.checkable = checkable
        self.shortcut = shortcut

    def __hash__(self):
        return hash((ToolbarAction, self.name))

    def __eq__(self, other):
        return isinstance(other, ToolbarAction) and self.name == other.name


class ToolbarSplitter(ToolbarAction):
    def __init__(self):
        super().__init__(None, None, None, None)



from typing import Optional

from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMenu


class MenuEntry:
    def __init__(self, caption, action, shortcut=None, checkable=False, checked=False, enabled=True, key=None):
        self.caption = caption
        self.action = action
        self.shortcut = shortcut
        self.checkable = checkable
        self.checked_initially = checked
        self.enabled = enabled
        self.key = key

        self._qaction = None

    def set_qaction(self, qaction: QAction):
        self._qaction = qaction
        self._enable(self.enabled)

    def enable(self):
        self._enable(True)

    def disable(self):
        self._enable(False)

    def _enable(self, b: bool):
        self.enabled = b
        if self._qaction is not None:
            self._qaction.setEnabled(b)

    @property
    def checked(self):
        if self._qaction is None or not self.checkable:
            return False
        return self._qaction.isChecked()

    @checked.setter
    def checked(self, checked: bool):
        if self._qaction is not None:
            self._qaction.setChecked(checked)


class MenuSeparator:
    def __init__(self):
        pass


class Menu:
    def __init__(self, caption, children=(), parent=None):
        self.parent = parent
        self.caption = caption

        self.entries = []
        self._keyed_entries = None

        self._qmenu: Optional[QMenu] = None

        for child in children:
            self.add(child)

    def action_by_key(self, key):
        if not self._keyed_entries:
            self._keyed_entries = {ent.key: ent for ent in self.entries if isinstance(ent, MenuEntry)}
        return self._keyed_entries.get(key, None)

    def qmenu(self, extra_entries=None, cached=True):
        if extra_entries is None:
            extra_entries = []

        if cached and not extra_entries and self._qmenu is not None:
            # in order to use the cached result, must not have extra entries
            return self._qmenu

        menu = QMenu(self.caption, self.parent) if self.parent is not None else QMenu(self.caption)

        for entry in self.entries + extra_entries:
            self.translate_element(menu, entry)

        # in order to cache the result, must not have extra entries
        if not extra_entries:
            self._qmenu = menu

        return menu

    @staticmethod
    def translate_element(menu, entry, index=None):
        if index is None:
            before = None
        else:
            try:
                before = menu.actions()[index]
            except IndexError:
                before = None

        if entry is None:
            entry = MenuSeparator()
        elif type(entry) is tuple and len(entry) == 2 and callable(entry[1]):
            entry = MenuEntry(*entry)
        elif type(entry) is tuple and len(entry) == 2 and hasattr(entry[1], "__iter__"):
            entry = Menu(*entry)

        if isinstance(entry, MenuEntry):
            action = QAction(entry.caption, menu)
            action.triggered.connect(entry.action)
            entry.set_qaction(action)

            if entry.shortcut is not None:
                action.setShortcut(entry.shortcut)
            if entry.checkable:
                action.setCheckable(True)
                action.setChecked(entry.checked_initially)

            if before is None:
                menu.addAction(action)
            else:
                menu.insertAction(before, action)
        elif isinstance(entry, MenuSeparator):
            if before is None:
                menu.addSeparator()
            else:
                menu.insertSeparator(before)
        elif isinstance(entry, Menu):
            if before is None:
                menu.addMenu(entry.qmenu())
            else:
                menu.insertMenu(before, entry.qmenu())
        elif isinstance(entry, QMenu):
            if before is None:
                menu.addMenu(entry)
            else:
                menu.insertMenu(before, entry)
        elif isinstance(entry, QAction):
            if before is None:
                menu.addAction(entry)
            else:
                menu.insertAction(before, entry)
        else:
            raise TypeError("Unsupported type", type(entry))

    def add(self, action, index=None):
        if index is None:
            index = len(self.entries)
        self.entries.insert(index, action)
        if self._qmenu is not None:
            self.translate_element(self._qmenu, action, index)

    def remove(self, action):
        self.entries.remove(action)
        if self._qmenu is not None and isinstance(action, MenuEntry):
            self._qmenu.removeAction(action._qaction)



from typing import TYPE_CHECKING, Callable, Optional

from angrmanagement.data.object_container import EventSentinel, ObjectContainer

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace


class Debugger:
    """
    Provides a generic interface with common debugger operations to control program execution and inspect program state.
    """

    def __init__(self, workspace: "Workspace"):
        super().__init__()
        self.workspace: Workspace = workspace
        self.instance: Instance = workspace.main_instance
        self.state_changed: EventSentinel = EventSentinel()

    @property
    def state_description(self) -> str:
        """
        Get a string describing the current debugging state.
        """
        return ""

    def init(self):
        """
        Initialize target connection.
        """

    @property
    def is_running(self) -> bool:
        """
        Determine if target is running.
        """
        return False

    @property
    def can_step_backward(self) -> bool:
        """
        Determine if the target can step backward by one machine instruction.
        """
        return False

    def step_backward(self):
        """
        Step backward by one machine instruction.
        """
        raise NotImplementedError

    @property
    def can_step_forward(self) -> bool:
        """
        Determine if the target can step forward by one machine instruction.
        """
        return False

    def step_forward(self, until_addr: Optional[int] = None):
        """
        Step forward by one machine instruction.
        """
        raise NotImplementedError

    @property
    def can_continue_backward(self) -> bool:
        """
        Determine if execution can continue in reverse.
        """
        return False

    def continue_backward(self):
        """
        Continue execution in reverse.
        """
        raise NotImplementedError

    @property
    def can_continue_forward(self) -> bool:
        """
        Determine if execution can continue.
        """
        return False

    def continue_forward(self):
        """
        Continue execution.
        """
        raise NotImplementedError

    @property
    def can_halt(self) -> bool:
        """
        Determine if the target can be interrupted.
        """
        return False

    def halt(self):
        """
        Interrupt the target.
        """
        raise NotImplementedError

    @property
    def is_halted(self) -> bool:
        """
        Determine if the target has been interrupted and is now halted.
        """
        return False

    @property
    def can_stop(self) -> bool:
        """
        Determine if the target can be stopped.
        """
        return False

    def stop(self):
        """
        Stop the target.
        """
        raise NotImplementedError

    @property
    def is_exited(self) -> bool:
        """
        Determine if the target has exited.
        """
        return False


class DebuggerListManager:
    """
    Manages the list of active debuggers.
    """

    def __init__(self):
        self.debugger_list = ObjectContainer([], "List of active debuggers")

    def add_debugger(self, dbg: Debugger):
        self.debugger_list.append(dbg)
        self.debugger_list.am_event(added=dbg)

    def remove_debugger(self, dbg: Debugger):
        self.debugger_list.remove(dbg)
        self.debugger_list.am_event(removed=dbg)


class DebuggerManager:
    """
    Manages one selected active debugger container.
    """

    def __init__(self, debugger_list_mgr: DebuggerListManager):
        self.debugger: ObjectContainer = ObjectContainer(None, "Current debugger")
        debugger_list_mgr.debugger_list.am_subscribe(self._on_debugger_list_event)

    def _on_debugger_list_event(self, **kwargs):
        if "removed" in kwargs:
            self._on_debugger_removed(kwargs["removed"])

    def _on_debugger_removed(self, dbg: Debugger):
        if self.debugger.am_obj is dbg:
            self.set_debugger(None)

    def set_debugger(self, dbg: Optional[Debugger]):
        self.debugger.am_obj = dbg
        self.debugger.am_event()


class DebuggerWatcher:
    """
    Watcher object that subscribes to debugger events whenever debugger changes.
    """

    def __init__(self, state_updated_callback: Callable, debugger: ObjectContainer):
        """
        :param state_updated_callback: Callable to be called whenever the debugger state changes.
        :param debugger: Debugger container to monitor.
        """
        super().__init__()
        self._last_selected_debugger: Optional[Debugger] = None
        self.state_updated_callback: Callable = state_updated_callback
        self.debugger: ObjectContainer = debugger
        self.debugger.am_subscribe(self._on_debugger_updated)
        self._subscribe_to_events()

    def shutdown(self):
        self.debugger.am_unsubscribe(self._on_debugger_updated)
        self._unsubscribe_from_events()

    def _unsubscribe_from_events(self):
        if self._last_selected_debugger:
            self._last_selected_debugger.state_changed.am_unsubscribe(self._on_debugger_state_updated)
            self._last_selected_debugger = None

    def _subscribe_to_events(self):
        if not self.debugger.am_none:
            self.debugger.state_changed.am_subscribe(self._on_debugger_state_updated)
            self._last_selected_debugger = self.debugger.am_obj

    def _on_debugger_updated(self, *args, **kwargs):  # pylint:disable=unused-argument
        self._unsubscribe_from_events()
        self._subscribe_to_events()
        self._on_debugger_state_updated()

    def _on_debugger_state_updated(self):
        self.state_updated_callback()
